"""Implementation of a vanilla transformer"""
from torch import nn


class VanillaTransformer(nn.Module):
    """A vanilla transformer implementation"""
    def __init__(self):
        super().__init__()

    def forward(self):
        pass
